from freighter import URL
from pydantic import BaseModel

from .auth import AuthenticationClient

from .channel import ChannelClient
from .channel.create import ChannelCreator
from .channel.registry import ChannelRegistry
from .channel.retrieve import ChannelRetriever
from .segment import SegmentClient

from .transport import Transport


class SynnaxOptions(BaseModel):
    """Options class for the Synnax client.
    """
    host: str
    port: int
    username: str = ""
    password: str = ""


class Synnax:
    """Client to perform operations against a Synnax cluster. If no credentials are provided
    in the options, the client will attempt to load them from the configuration file (
    ~/.synnax/config.json) or from environment variables.

    :param host: Hostname of a Synnax server.
    :param port: Port of a Synnax server.
    :param username: Username to authenticate with. Not required if the Synnax cluster
    is running in insecure mode.
    :param password: Password to authenticate with. Not required if the Synnax cluster
    is running in insecure mode.
    """

    _transport: Transport
    channel: ChannelClient
    data: SegmentClient

    def __init__(
        self,
        host: str = "",
        port: int = 0,
        username: str = "",
        password: str = "",
    ):
        self._transport = Transport(URL(host=host, port=port))
        if username != "" or password != "":
            auth = AuthenticationClient(
                self._transport.http.post_client(), username, password
            )
            auth.authenticate()
            self._transport.use(*auth.middleware())
        ch_retriever = ChannelRetriever(self._transport.http)
        ch_creator = ChannelCreator(self._transport.http)
        ch_registry = ChannelRegistry(ch_retriever)
        self.data = SegmentClient(self._transport, ch_registry)
        self.channel = ChannelClient(self.data, ch_retriever, ch_creator)



from enum import Enum

from freighter import EOF, ExceptionPayload, Payload, Stream, StreamClient

from synnax.channel.registry import ChannelRegistry
from synnax.telem import TimeRange, TimeSpan, TimeStamp

from .encoder import NumpyEncoderDecoder
from .payload import SegmentPayload
from .sugared import NumpySegment

AUTO_SPAN = TimeSpan(-1)


class _Command(int, Enum):
    OPEN = 0
    NEXT = 1
    PREV = 2
    SEEK_FIRST = 3
    SEEK_LAST = 4
    SEEK_LE = 5
    SEEK_GE = 6
    VALID = 7
    ERROR = 8


class _ResponseVariant(int, Enum):
    NONE = 0
    ACK = 1
    DATA = 2


class _Request(Payload):
    command: _Command
    span: TimeSpan | None = None
    range: TimeRange | None = None
    stamp: TimeStamp | None = None
    keys: list[str] | None = None


class _Response(Payload):
    variant: _ResponseVariant
    ack: bool
    command: _Command | None = None
    error: ExceptionPayload | None = None
    segments: list[SegmentPayload] | None = None


class CoreIterator:
    """Used to iterate over a databases telemetry in time-order. It should not be
    instantiated directly, and should instead be instantiated using the segment Client.

    Using an iterator is ideal when querying/processing large ranges of data, but is
    relatively complex and difficult to use. If you're looking to retrieve telemetry
    between two timestamps, see the segment Client read method instead.
    """

    _ENDPOINT = "/segment/iterate"

    client: StreamClient
    stream: Stream[_Request, _Response]
    values: list[SegmentPayload]
    aggregate: bool

    def __init__(self, client: StreamClient, aggregate: bool = False) -> None:
        self.client = client
        self.aggregate = aggregate

    def open(self, keys: list[str], tr: TimeRange):
        """Opens the iterator, configuring it to iterate over the telemetry in the
        channels with the given keys within the provided time range.

        :param keys: The keys of the channels to iterate over.
        :param tr: The time range to iterate over.
        """
        self.stream = self.client.stream(self._ENDPOINT, _Request, _Response)
        self._exec(command=_Command.OPEN, range=tr, keys=keys)
        self.values = []

    def next(self, span: TimeSpan) -> bool:
        """Reads the next time span of telemetry for each channel in the iterator.

        :param span: The span of time to read. A negative span is equivalent to calling
        prev with the absolute value of the span. If AUTO_SPAN is provided, the iterator
        will automatically determine the span of time to read. This is useful for iterating
        quickly over an entire time range.

        :returns: False if a segment satisfying the request can't be found for a
        particular channel or the iterator has accumulated an error.
        """
        return self._exec(command=_Command.NEXT, span=span)

    def prev(self, span: TimeSpan) -> bool:
        """Reads the previous time span of telemetry for each channel in the iterator.

        :param span: The span of time to read. A negative span is equivalent to calling
        next with the absolute value of the span. If AUTO_SPAN is provided, the iterator
        will automatically determine the span of time to read. This is useful for iterating
        quickly over an entire time range.

        :returns: False if a segment satisfying the request can't be found for a particular
        channel or the iterator has accumulated an error.
        """
        return self._exec(command=_Command.NEXT, span=span)

    def seek_first(self) -> bool:
        """Seeks the iterator to the first segment in the time range, but does not read
        it. Also invalidates the iterator. The iterator will not be considered valid
        until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_FIRST)

    def seek_last(self) -> bool:
        """Seeks the iterator to the last segment in the time range, but does not read it.
        Also invalidates the iterator. The iterator will not be considered valid
        until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_LAST)

    def seek_lt(self, stamp: TimeStamp) -> bool:
        """Seeks the iterator to the first segment whose start is less than or equal to
        the provided timestamp. Also invalidates the iterator. The iterator will not be
        considered valid until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_LE, stamp=stamp)

    def seek_ge(self, stamp: TimeStamp) -> bool:
        """Seeks the iterator to the first segment whose start is greater than or equal to
        the provided timestamp. Also invalidates the iterator. The iterator will not be
        considered valid until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_GE, stamp=stamp)

    def valid(self) -> bool:
        """Returns true if the iterator value contains a valid segment, and False otherwise.
        valid most commonly returns false when the iterator is exhausted or has accumulated
        an error.
        """
        return self._exec(command=_Command.VALID)

    def close(self):
        """Close closes the iterator. An iterator MUST be closed after use, and this method
        should probably be placed in a 'finally' block. If the iterator is not closed, it make
        leak resources and threads.
        """
        exc = self.stream.close_send()
        if exc is not None:
            raise exc
        pld, exc = self.stream.receive()
        if not isinstance(exc, EOF):
            raise exc

    def _exec(self, **kwargs) -> bool:
        exc = self.stream.send(_Request(**kwargs))
        if exc is not None:
            raise exc
        if not self.aggregate:
            self.values = []
        while True:
            r, exc = self.stream.receive()
            if exc is not None:
                raise exc
            if r.variant == _ResponseVariant.ACK:
                return r.ack
            self.values += r.segments


class NumpyIterator(CoreIterator):
    """Used to iterate over a databases telemetry in time-order. It should not be
    instantiated directly, and should instead be instantiated using the segment Client.

    Using an iterator is ideal when querying/processing large ranges of data, but is
    relatively complex and difficult to use. If you're looking to retrieve telemetry
    between two timestamps, see the segment Client read method instead.
    """

    decoder: NumpyEncoderDecoder
    channels: ChannelRegistry

    def __init__(
        self,
        transport: StreamClient,
        channels: ChannelRegistry,
        aggregate: bool = False,
    ):
        super().__init__(transport, aggregate)
        self.decoder = NumpyEncoderDecoder()
        self.channels = channels

    def open(self, keys: list[str], tr: TimeRange) -> None:
        super().open(keys, tr)

    @property
    def value(self) -> dict[str, NumpySegment]:
        """
        :returns: The current iterator value as a dictionary whose keys are channels
        and values are segments containing telemetry at the current iterator position.
        """
        decoded = []
        self.values.sort(key=lambda v: v.start)
        res = dict()
        for i, seg in enumerate(self.values):
            decoded.append(self.decoder.decode(self.channels.get(seg.channel_key), seg))
        for i, dec in enumerate(decoded):
            if dec.channel.key in res:
                res[dec.channel.key].extend(dec)
            else:
                res[dec.channel.key] = dec
        return res



from .. import telem
from .sugared import SugaredBinarySegment


class Splitter:
    threshold: telem.Size

    def __init__(self, threshold: telem.Size) -> None:
        self.threshold = threshold

    def split(self, segment: SugaredBinarySegment) -> list[SugaredBinarySegment]:
        if segment.size <= self.threshold:
            return [segment]
        split_v = self.threshold - (self.threshold % segment.channel.density)
        truncated = SugaredBinarySegment(
            segment.channel, segment.start, segment.data[:split_v]
        )
        _next = SugaredBinarySegment(
            segment.channel, truncated.end, segment.data[split_v:]
        )
        return [truncated, *self.split(_next)]



import keyring
from .file import ConfigFile
from pydantic import BaseModel

from synnax import SynnaxOptions


class ClusterConfig(BaseModel):
    options: SynnaxOptions


class ClustersConfig:
    internal: ConfigFile

    def __init__(self, config: ConfigFile):
        self.internal = config

    def get(self, key: str = "default") -> ClusterConfig | None:
        c = self.internal.get(f"clusters.{key}")
        if c is not None:
            c["options"]["password"] = keyring.get_password("synnax", key)
        return ClusterConfig(**c)

    def set(self, c: ClusterConfig, key: str = "default"):
        p = c.dict()
        keyring.set_password("synnax", key, p["options"].pop("password"))
        self.internal.set(f"clusters.{key}", p)



import os
from pathlib import Path

from freighter.exceptions import Unreachable

from .. import Synnax, AuthError, SynnaxOptions
from .flow import Context
from ..config import ConfigFile, ClustersConfig


def load_config_options(ctx: Context) -> SynnaxOptions:
    """Loads the connection parameters from a configuration file.

    :param ctx: The context of the current flow.
    :return: The options to connect to a Synnax server.
    """
    cfg = ClustersConfig(ConfigFile(Path(os.path.expanduser("~/.synnax"))))
    return cfg.get().options


def prompt_client_options(ctx: Context) -> SynnaxOptions:
    """Prompts the user for the parameters to connect to a Synnax server.

    :param ctx: The context of the current flow.
    :return: The options to connect to a Synnax server.
    """
    ctx.console.info("Enter your Synnax connection parameters:")
    params = dict()
    params['host'] = ctx.console.ask("Host", default="localhost")
    params['port'] = ctx.console.ask_int("Port", default=9090)
    params['username'] = ctx.console.ask("Username", default="synnax")
    params['password'] = ctx.console.ask_password("Password")
    return SynnaxOptions(**params)


def connect_client(ctx: Context, opts: SynnaxOptions) -> Synnax | None:
    """Connects to a Synnax server. Prints user-friendly messages to the console if
    the connection fails.

    :param ctx: The context of the current flow.
    :param opts: The options to connect to a Synnax server.
    :return: The connected Synnax client, or None if the connection failed.
    """
    try:
        client = Synnax(**opts.dict())
    except Unreachable:
        return ctx.console.error(
            f"Cannot reach Synnax server at {opts.host}:{opts.port}"
        )
    except AuthError:
        return ctx.console.error("Invalid credentials")
    except Exception as e:
        raise e
        # return ctx.console.error(f"An error occurred: {e}")
    ctx.console.success("Connection successful!")
    return client



from freighter import (
    URL,
    AsyncStreamClient,
    HTTPClientFactory,
    JSONEncoder,
    MsgpackEncoder,
    StreamClient,
    SyncStreamClient,
    WebsocketClient,
    Middleware,
    AsyncMiddleware,
)


class Transport:
    url: URL
    stream: StreamClient
    stream_async: AsyncStreamClient
    http: HTTPClientFactory

    def __init__(self, url: URL) -> None:
        self.url = url.child("/api/v1/")
        self.stream_async = WebsocketClient(
            base_url=self.url, encoder=MsgpackEncoder(), max_message_size=int(5e6)
        )
        self.stream = SyncStreamClient(self.stream_async)
        self.http = HTTPClientFactory(url=self.url, encoder_decoder=JSONEncoder())

    def use(self, *middleware: Middleware):
        self.http.use(*middleware)
        self.stream.use(*middleware)

    def use_async(self, *middleware: AsyncMiddleware):
        self.http.use(*middleware)
        self.stream_async.use(*middleware)



from typing import Any

from pydantic import BaseModel


class ChannelMeta(BaseModel):
    """General channel metadata that can be read from a file.
    """
    name: str
    """The name of the channel."""
    meta_data: dict[str, Any]
    """Any additional metadata associated with the channel."""



from .payload import ChannelPayload
from .retrieve import ChannelRetriever


class ChannelRegistry:
    retriever: ChannelRetriever
    channels: dict[str, ChannelPayload]

    def __init__(self, retriever: ChannelRetriever, channels=None) -> None:
        if channels is None:
            channels = list()
        self.retriever = retriever
        self.channels = {ch.name: ch for ch in channels}

    def get(self, key: str) -> ChannelPayload | None:
        record = self.channels.get(key, None)
        if record is None:
            record = self.retriever.get(key=key)
            self.channels[key] = record
        return record

    def get_n(self, keys: list[str]) -> list[ChannelPayload]:
        results = list()
        retrieve_keys = list()
        for key in keys:
            record = self.channels.get(key, None)
            if record is not None:
                results.append(record)
            retrieve_keys.append(key)
        if retrieve_keys:
            results.extend(self.retriever.retrieve(retrieve_keys))
        return results



__version__ = "0.1.0"

from .channel import Channel
from .synnax import Synnax, SynnaxOptions
from .exceptions import (
    AuthError,
    ContiguityError,
    GeneralError,
    ParseError,
    QueryError,
    RouteError,
    UnexpectedError,
    ValidationError,
    Field,
)
from .telem import (
    BIT8,
    BIT16,
    BIT32,
    BIT64,
    DATA_TYPE_UNKNOWN,
    FLOAT32,
    FLOAT64,
    HOUR,
    HZ,
    INT8,
    INT16,
    INT32,
    INT64,
    KHZ,
    MHZ,
    MICROSECOND,
    MILLISECOND,
    MINUTE,
    NANOSECOND,
    SECOND,
    TIME_RANGE_MAX,
    TIME_SPAN_MAX,
    TIME_STAMP_MAX,
    TIME_STAMP_MIN,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
    DataType,
    Density,
    Rate,
    Size,
    TimeRange,
    TimeSpan,
    TimeStamp,
    UnparsedDataType,
    UnparsedDensity,
    UnparsedRate,
    UnparsedTimeSpan,
    UnparsedTimeStamp,
    now,
    since,
)



from typing import Generic, TypeVar, Callable

from synnax.cli.console import Console

T = TypeVar('T')


class Context:
    console: Console

    def __init__(self, console: Console):
        self.console = console


class Flow(Generic[T]):
    steps: dict[str, Callable[[Context, T], str | None]]
    context: Context

    def __init__(self, ctx: Context):
        self.steps = {}
        self.context = ctx

    def add(self, name: str, step: T):
        self.steps[name] = step

    def run(self, req: T, root: str):
        root_step = self.steps[root]
        self._run(root_step, req)

    def _run(self, step: Callable[[T, Context], str | None], request: T, ):
        next_step = step(self.context, request)
        if next_step is not None:
            self._run(self.steps[next_step], request)



from __future__ import annotations

from dataclasses import dataclass

from numpy import append, ndarray

from synnax.channel.payload import ChannelPayload
from synnax.exceptions import ContiguityError, ValidationError
from synnax.telem import Size, TimeRange, TimeSpan, TimeStamp, UnparsedTimeStamp

from .payload import SegmentPayload


@dataclass
class SugaredHeader:
    channel: ChannelPayload
    start: TimeStamp = TimeStamp(0)

    def __init__(self, channel: ChannelPayload, start: UnparsedTimeStamp):
        self.channel = channel
        self.start = TimeStamp(start)


class SugaredBinarySegment(SugaredHeader):
    data: bytes = b""

    def __init__(
        self, channel: ChannelPayload, start: UnparsedTimeStamp, data: bytes = b""
    ):
        super().__init__(channel, start)
        self.data = data

    def payload(self) -> SegmentPayload:
        return SegmentPayload(
            channel_key=self.channel.key, start=self.start, data=self.data
        )

    @classmethod
    def sugar(cls, ch: ChannelPayload, seg: SegmentPayload) -> SugaredBinarySegment:
        return cls(ch, seg.start, seg.data)

    @property
    def size(self) -> Size:
        return Size(len(self.data))

    @property
    def span(self) -> TimeSpan:
        return self.channel.rate.size_span(self.size, self.channel.density)

    @property
    def range(self) -> TimeRange:
        return self.start.span_range(self.span)

    @property
    def end(self) -> TimeStamp:
        return self.range.end


@dataclass
class NumpySegment(SugaredHeader):
    data: ndarray = ndarray([])

    def __init__(self, channel: ChannelPayload, start: TimeStamp, data: ndarray):
        super().__init__(channel, start)
        self.data = data

    @property
    def span(self) -> TimeSpan:
        return self.channel.rate.span(self.data.size)

    @property
    def range(self) -> TimeRange:
        return self.start.span_range(self.span)

    @property
    def end(self) -> TimeStamp:
        return self.range.end

    def extend(self, other: NumpySegment):
        if self.channel.key != other.channel.key:
            raise ValidationError(
                f"""Cannot extend segment because channel keys mismatch.
                Segment Channel: {self.channel.key}
                Next Segment Channel: {other.channel.key}
                """
            )
        if self.end != other.start:
            raise ContiguityError(
                f"""Cannot extend segment because end and start times are not equal.
                Segment End: {self.end}
                Next Segment Start: {other.start}
                """
            )
        self.data = append(self.data, other.data)



from numpy import ndarray

from synnax.segment.client import SegmentClient as SegmentClient
from synnax.telem import (
    DATA_TYPE_UNKNOWN,
    Rate,
    UnparsedDataType,
    UnparsedRate,
    UnparsedTimeStamp,
    UnparsedDensity,
)

from .create import ChannelCreator
from .payload import ChannelPayload
from .retrieve import ChannelRetriever


class Channel(ChannelPayload):
    """Represents a Channel in a Synnax database.
    """

    segment_client: SegmentClient | None = None

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self,
        data_type: UnparsedDataType,
        rate: UnparsedRate,
        name: str = "",
        node_id: int = 0,
        key: str = "",
        density: UnparsedDensity = 0,
        is_index: bool = False,
        index: str = "",
        segment_client: SegmentClient = None,
    ):
        super().__init__(
            data_type=data_type,
            rate=rate,
            name=name,
            node_id=node_id,
            key=key,
            density=density,
            is_index=is_index,
            index=index,
        )
        self.segment_client = segment_client

    def _payload(self) -> ChannelPayload:
        return ChannelPayload(
            data_type=self.data_type,
            density=self.density,
            rate=self.rate,
            name=self.name,
            node_id=self.node_id,
            key=self.key,
            index=self.index,
            is_index=self.is_index,
        )

    def read(self, start: UnparsedTimeStamp, end: UnparsedTimeStamp) -> ndarray:
        """Reads telemetry from the channel between the two timestamps.

        :param start: The starting timestamp of the range to read from.
        :param end: The ending timestamp of the range to read from.
        :returns: A numpy array containing the retrieved telemetry from the database.
        :raises ContiguityError: If the telemetry between start and end is non-contiguous.
        """
        return self.segment_client.read(self.key, start, end)

    def write(self, start: UnparsedTimeStamp, data: ndarray):
        """Writes telemetry to the channel starting at the given timestamp.

        :param start: The starting timestamp of the first sample in data.
        :param data: The telemetry to write to the channel.
        :returns: None.
        """
        self.segment_client.write(self.key, start, data)

    def _assert_created(self):
        if not self.segment_client:
            raise ValidationError("cannot read from a channel that has not been created")

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return self.key == other.key


class ChannelClient:
    """The core client class for executing channel operations against a Synnax cluster."""

    _segment_client: SegmentClient
    _retriever: ChannelRetriever
    _creator: ChannelCreator

    def __init__(
        self,
        segment_client: SegmentClient,
        retriever: ChannelRetriever,
        creator: ChannelCreator,
    ):
        self._segment_client = segment_client
        self._retriever = retriever
        self._creator = creator

    def create_many(self, channels: list[Channel]) -> list[Channel]:
        """Creates all channels in the given list.
        """
        return self._sugar(*self._creator.create_many([c._payload() for c in channels]))

    def create(
        self,
        name: str = "",
        node_id: int = 0,
        rate: UnparsedRate = Rate(0),
        data_type: UnparsedDataType = DATA_TYPE_UNKNOWN,
        index: str = "",
        is_index: bool = False,
    ) -> Channel:
        """Creates a channel using the given template.

        :param name: The name of the channel to create.
        :param rate: The sample rate of the channel in Hz.
        :param data_type: The data type of the channel. Can be any type in
        UnparsedDataType, such as np.float64 or np.int64,
        :param node_id: The node that holds the lease on the channel. If you don't know
        what this is, don't worry about it.
        :returns: The created channel.
        """
        return self._sugar(self._creator.create(name=name, node_id=node_id, rate=rate, data_type=data_type, index=index, is_index=is_index))[0]

    def get(self, key: str = None, name: str = None) -> Channel:
        """Retrieves channels with the given keys.

        :param keys: The list of keys to retrieve channels for.
        :raises QueryError: If any of the channels can't be found.
        :returns: A list of retrieved Channels.
        """
        return self._sugar(self._retriever.get(key, name))[0]

    def filter(
        self,
        keys: list[str] = None,
        names: list[str] = None,
        node_id: int = None,
    ) -> list[Channel]:
        """Filters channels using the given parameters.

        :param kwargs: The parameters to filter channels by.
        :returns: A list of channels that match the given parameters.
        """
        return self._sugar(*self._retriever.filter(keys, names, node_id))

    def _sugar(self, *channels: ChannelPayload) -> list[Channel]:
        return [Channel(**c.dict(), segment_client=self._segment_client) for c in channels]



from typing import Callable
from freighter import (
    UnaryClient,
    Payload,
    MetaData,
    Next,
    Middleware,
    AsyncMiddleware,
    AsyncNext,
)

from synnax.user.payload import UserPayload


class InsecureCredentials(Payload):
    username: str
    password: str


class TokenResponse(Payload):
    token: str
    user: UserPayload


AUTHORIZATION_HEADER = "Authorization"
TOKEN_REFRESH_HEADER = "Refresh-Token"


def token_middleware(token: Callable[[], str],
                     set_token: Callable[[str], None]) -> Middleware:
    def mw(md: MetaData, _next: Next):
        md.set(AUTHORIZATION_HEADER, "Bearer " + token())
        out_md, exc = _next(md)
        if TOKEN_REFRESH_HEADER in out_md.params:
            tk = out_md.get(TOKEN_REFRESH_HEADER)
            set_token(tk)
        return out_md, exc

    return mw


def async_token_middleware(token: Callable[[], str],
                           set_token: Callable[[str], None]) -> Middleware:
    async def mw(md: MetaData, _next: AsyncNext):
        md.set(AUTHORIZATION_HEADER, "Bearer " + token())
        out_md, exc = await _next(md)
        if TOKEN_REFRESH_HEADER in out_md:
            token = out_md.get(TOKEN_REFRESH_HEADER)
            set_token(token)
        return out_md, exc

    return mw


class AuthenticationClient:
    _ENDPOINT = "/auth/login"

    client: UnaryClient
    username: str
    password: str
    token: str
    user: UserPayload

    def __init__(
        self,
        transport: UnaryClient,
        username: str,
        password: str,
    ) -> None:
        self.client = transport
        self.username = username
        self.password = password

    def authenticate(self) -> None:
        res, exc = self.client.send(
            self._ENDPOINT,
            InsecureCredentials(username=self.username, password=self.password),
            TokenResponse,
        )
        if exc is not None:
            raise exc
        self.token = res.token
        self.user = res.user

    def get_token(self) -> str:
        return self.token

    def set_token(self, token: str) -> None:
        self.token = token

    def middleware(self) -> list[Middleware]:
        return [token_middleware(self.get_token, self.set_token)]

    def async_middleware(self) -> list[AsyncMiddleware]:
        return [async_token_middleware(self.get_token, self.set_token)]



from freighter import Payload

from synnax.telem import DataType, Density, Rate


class ChannelPayload(Payload):
    """A payload container that represent the properties of a channel exchanged to and
    from the Synnax server.
    """

    data_type: DataType
    rate: Rate = 0
    density: Density = Density(0)
    name: str = ""
    node_id: int = 0
    key: str = ""
    is_index: bool = False
    index: str = ""



from rich import print
from rich.prompt import Prompt, Confirm, FloatPrompt, IntPrompt


class RichConsole:
    """A rich text implementation of the Console protocol.
    """
    info_color: str
    warn_color: str
    error_color: str
    success_color: str

    def __init__(
        self,
        info_color: str = "#3774D0",
        warn_color: str = "yellow",
        error_color: str = "red",
        success_color: str = "green",
    ):
        self.info_color = info_color
        self.warn_color = warn_color
        self.error_color = error_color
        self.success_color = success_color

    def info(self, message: str) -> None:
        print(f"[{self.info_color}]{message}[/]")

    def warn(self, message: str) -> None:
        print(f"[{self.warn_color}]{message}[/]")

    def error(self, message: str) -> None:
        print(f"[{self.error_color}]{message}[/]")

    def success(self, message: str) -> None:
        print(f"[{self.success_color}]{message}[/]")

    def ask(
        self,
        question: str,
        choices: list[str] = None,
        default: str = None,
    ) -> str:
        return Prompt.ask(
            question,
            choices=choices,
            default=default,
        )

    def ask_int(
        self,
        question: str,
        default: int = None,
    ) -> int:
        return IntPrompt.ask(
            question,
            default=default,
        )

    def ask_float(
        self,
        question: str,
        default: float = None,
    ) -> float:
        return FloatPrompt.ask(
            question,
            default=default,
        )

    def ask_password(
        self,
        question: str,
    ) -> str:
        return Prompt.ask(
            question,
            password=True,
        )

    def confirm(
        self,
        question: str,
        default: bool = None,
    ) -> bool:
        return Confirm.ask(
            question,
            default=default,
        )

    def table(
        self,
        columns: list[str],
        rows: list[dict],
    ) -> None:
        from rich.table import Table

        table = Table(show_header=True, header_style="bold magenta")
        for column in columns:
            table.add_column(column)
        for row in rows:
            # order the row.py by the columns
            table.add_row(*[row[column] for column in columns])

        print(table)



from .protocol import RowReader, Writer, Matcher
from .csv import CSVReader, CSVWriter
from pathlib import Path

READERS: list[type[RowReader]] = [
    CSVReader,
]

WRITERS: list[type[Writer]] = [
    CSVWriter,
]


class IOFactory:
    """A registry for retrieving readers for different file types.
    """
    reader_classes: list[type[RowReader]]
    writer_classes: list[type[Writer]]

    def __init__(
        self,
        readers: list[type[RowReader]] = None,
        writers: list[type[Writer]] = None,
    ):
        self.reader_classes = readers or READERS
        self.writer_classes = writers or WRITERS

    def new_reader(self, path: Path) -> RowReader:
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")

        if not path.is_file():
            raise IsADirectoryError(f"Path is a directory: {path}")

        for _Reader in self.reader_classes:
            if _Reader.match(path):
                return _Reader(path)

        raise NotImplementedError(f"File type not supported: {path}")

    def new_writer(self, path: Path) -> Writer:
        if not path.parent.exists():
            raise FileNotFoundError(f"File not found: {path}")

        if not path.parent.is_dir():
            raise IsADirectoryError(f"Path is a directory: {path}")

        for _Writer in self.writer_classes:
            if _Writer.match(path):
                return _Writer(path)

        raise NotImplementedError(f"File type not supported: {path}")

    def extensions(self) -> list[str]:
        extensions = set()
        for reader in self.reader_classes:
            extensions.update(reader.extensions())
        return list(extensions)



from pathlib import Path

import pandas as pd
from pandas.io.parsers import TextFileReader

from synnax.io.matcher import new_extension_matcher
from synnax.io.protocol import ChannelMeta, ReaderType

CSVMatcher = new_extension_matcher(["csv"])


class CSVReader(CSVMatcher):
    """A RowReader implementation for CSV files.
    """
    reader: TextFileReader
    _path: Path
    _channels: list[ChannelMeta] | None
    channel_keys: list[str] | None

    def __init__(self,
                 path: Path,
                 channel_keys: list[str] = None,
                 chunk_size: int = None,
                 ):
        self._path = path
        self.channel_keys = channel_keys
        self._channels = None

    def channels(self) -> list[ChannelMeta]:
        if not self._channels:
            self._channels = [ChannelMeta(name=name, meta_data={}) for name in
                              pd.read_csv(self._path, nrows=0).columns]
        return self._channels

    def set_chunk_size(self, chunk_size: int):
        self.reader = pd.read_csv(
            self._path,
            chunksize=chunk_size,
            usecols=self.channel_keys,
        )

    def read(self) -> pd.DataFrame:
        return next(self.reader)

    @classmethod
    def type(cls) -> ReaderType:
        return ReaderType.Row

    def path(self) -> Path:
        return self._path


class CSVWriter(CSVMatcher):
    _path: Path

    def __init__(
        self,
        path: Path,
    ):
        self._path = path

    def write(self, df: pd.DataFrame):
        df.to_csv(self._path, index=False)

    def path(self) -> Path:
        return self._path



import gc
from datetime import datetime

from pandas import DataFrame

from .. import Channel, Synnax
from ..telem import MEGABYTE
from ..io import RowReader
from ..segment import NumpyWriter


class RowIngestionEngine:
    """An ingestion engine that reads data from a row-based reader and writes it to a
    Synnax cluster.
    """
    client: Synnax
    writer: NumpyWriter
    reader: RowReader
    channels: list[Channel]
    idx_grouped: dict[Channel, list[Channel]]

    def __init__(
        self,
        client: Synnax,
        reader: RowReader,
        channels: list[Channel],
        soft_mem_limit: int = 500 * MEGABYTE,
    ):
        self.channels = channels
        self.idx_grouped = {ch: list() for ch in channels if ch.is_index}
        for ch in self.idx_grouped:
            self.idx_grouped[ch] = [_ch for _ch in channels if _ch.index == ch.key[-1]]

        self.mem_limit = soft_mem_limit
        self.reader = reader
        self.reader.set_chunk_size(self.get_chunk_size())
        self.client = client
        self.writer = self.client.data.new_writer([ch.key for ch in channels])

    def get_chunk_size(self):
        """Sum the density of all channels to determine the chunk size.
        """
        return self.mem_limit // sum(ch.density for ch in self.channels)

    def run(self):
        """Run the ingestion engine.
        """
        try:
            while True:
                try:
                    chunk = self.reader.read()
                    self._run_chunk(chunk)
                    gc.collect()
                except StopIteration:
                    break
            self.writer.commit()
        finally:
            self.writer.close()

    def _run_chunk(self, chunk: DataFrame):
        """Ingest a chunk of data.
        """
        for idx, channels in self.idx_grouped.items():
            idx_data = chunk[idx.name].to_numpy(dtype=idx.data_type)
            start = idx_data[0]
            print(self.writer.write(idx.key, start, idx_data))
            for ch in channels:
                t0 = datetime.now()
                ch_data = chunk[ch.name].to_numpy(dtype=ch.data_type)
                print("conversion time:", datetime.now() - t0)
                print(ch.key, len(ch_data))
                t0 = datetime.now()
                print(self.writer.write(ch.key, start, ch_data))
                print("write time:", datetime.now() - t0)



from __future__ import annotations
import typing

from freighter import HTTPClientFactory, Payload, UnaryClient

from .payload import ChannelPayload
from ..exceptions import ValidationError, QueryError


class _Request(Payload):
    keys: list[str] | None = None
    node_id: int | None = None
    names: list[str] | None = None


class _Response(Payload):
    channels: list[ChannelPayload] | None = []


class ChannelRetriever:
    _ENDPOINT = "/channel/retrieve"
    client: UnaryClient

    def __init__(self, client: HTTPClientFactory):
        self.client = client.get_client()

    def get(self, key: str = None, name: str = None) -> ChannelPayload:
        req = _Request()
        if key is None and name is None:
            raise ValidationError("Must specify a key or name")
        if key is not None:
            req.keys = [key]
        if name is not None:
            req.names = [name]
        res = self._execute(req)
        if len(res) == 0:
            raise QueryError("channel not found")
        return res[0]

    def filter(
        self,
        keys: list[str] = None,
        names: list[str] = None,
        node_id: int = None,
    ) -> list[ChannelPayload]:
        return self._execute(_Request(keys=keys, names=names))

    def _execute(self, req: _Request) -> list[ChannelPayload]:
        res, exc = self.client.send(self._ENDPOINT, req, _Response)
        if exc is not None:
            raise exc
        assert res is not None
        return res.channels



from .flow import Context

from pathlib import Path

from ..io import IOFactory, RowReader

io_factory = IOFactory()


def prompt_file(ctx: Context) -> Path | None:
    """
    Prompts the user for a file path.
    """
    fp = Path(ctx.console.ask("File path"))
    if not fp.exists():
        ctx.console.error(f"File does not exist: {fp}")
        return None
    ctx.console.success(f"File found: {fp}")
    return fp


def prompt_new_reader(ctx: Context) -> RowReader | None:
    """
    Prompts the user for a file path and returns a new reader for that file.
    """
    fp = prompt_file(ctx)
    if fp is None:
        return None
    return io_factory.new_reader(fp)



import numpy as np

from synnax.channel.registry import ChannelRegistry
from synnax.exceptions import ContiguityError
from synnax.segment.iterator import NumpyIterator, AUTO_SPAN
from synnax.segment.writer import CoreWriter, NumpyWriter
from synnax.telem import TimeRange, UnparsedTimeStamp
from synnax.transport import Transport

from . import iterator
from .sugared import NumpySegment


class SegmentClient:
    """SegmentClient provides interfaces for reading and writing segmented
    telemetry from a Synnax Cluster. SegmentClient should not be instantiated
    directly, but rather used through the synnax.Synnax class.
    """

    _transport: Transport
    _channels: ChannelRegistry

    def __init__(self, transport: Transport, registry: ChannelRegistry):
        self._transport = transport
        self._channels = registry

    def new_writer(self, keys: list[str]) -> NumpyWriter:
        """Opens a new writer on the given channels.

        :param keys: A list of channel keys that the writer will write to. A writer
        cannot write to keys not provided in this list. See the NumpyWriter documentation
        for more.
        :returns: A NumpyWriter that can be used to write telemetry to the given channels.
        """
        core = CoreWriter(client=self._transport.stream)
        npw = NumpyWriter(core=core, channels=self._channels)
        npw.open(keys)
        return npw

    def new_iterator(
        self,
        keys: list[str],
        tr: TimeRange,
        aggregate: bool = False,
    ) -> NumpyIterator:
        """Opens a new iterator over the given channels within the provided time range.

        :param keys: A list of channel keys to iterator over.
        :param tr: A time range to iterate over.
        :param aggregate:  Whether to accumulate iteration results or reset them on every
        iterator method call.
        :returns: A NumpyIterator over the given channels within the provided time
        range. See the NumpyIterator documentation for more.
        """
        npi = iterator.NumpyIterator(
            transport=self._transport.stream,
            channels=self._channels,
            aggregate=aggregate,
        )
        npi.open(keys, tr)
        return npi

    def write(self, to: str, start: UnparsedTimeStamp, data: np.ndarray):
        """Writes telemetry to the given channel starting at the given timestamp.

        :param to: The key of the channel to write to.
        :param start: The starting timestamp of the first sample in data.
        :param data: The telemetry to write to the channel.
        :returns: None.
        """
        _writer = self.new_writer([to])
        try:
            _writer.write(to, start, data)
            _writer.commit()
        finally:
            _writer.close()

    def read(
        self, from_: str, start: UnparsedTimeStamp, end: UnparsedTimeStamp
    ) -> np.ndarray:
        """Reads telemetry from the channel between the two timestamps.

        :param from_: THe key of the channel to read from.
        :param start: The starting timestamp of the range to read from.
        :param end: The ending timestamp of the range to read from.
        :returns: A numpy array containing the retrieved telemetry.
        :raises ContiguityError: If the telemetry between start and end is non-contiguous.
        """
        return self.read_segment(from_, start, end).data

    def read_segment(
        self, from_: str, start: UnparsedTimeStamp, end: UnparsedTimeStamp
    ) -> NumpySegment:
        """Reads a Segment from the given channel between the two timestamps.

        :param from_: The key of the channel to read from.
        :param start: The starting timestamp of the range to read from.
        :param end: The ending timestamp of the range to read from.
        :returns: A NumpySegment containing the read telemetry.
        :raises ContiguityError: If the telemetry between start and end is non-contiguous.
        """
        _iterator = self.new_iterator([from_], TimeRange(start, end), aggregate=True)
        seg = None
        try:
            _iterator.seek_first()
            while _iterator.next(AUTO_SPAN):
                pass
            seg = _iterator.value[from_]
        except ContiguityError as e:
            raise ContiguityError(
                f"""Channel data between {start} and {end} is non-contiguous.
                """
            ) from e
        finally:
            _iterator.close()
        return seg



from .flow import Context


def select_from_table(
    ctx: Context,
    columns: list[str],
    rows: list[dict[str, any]],
    allow_none: bool = False,
    default: int = None,
) -> int | None:
    """Prompts the user to select a row.py from a table.

    :param ctx: The current flow Context.
    :param columns: The columns of the table.
    :param rows: The rows of the table.
    :param allow_none: Whether to allow the user to select nothing.
    :param default: The default option to select. If a default is provided,
    allow_none is ignored.
    :returns: The index of the selected row.py or None if nothing was selected.
    """
    ctx.console.table(columns=["option", *columns], rows=[
        {"option": str(i), **row} for i, row in enumerate(rows)
    ])
    choices = [str(i) for i in range(len(rows))]
    if allow_none and default is None:
        default = "None"
        ctx.console.info("Press enter to select nothing.")
    i = ctx.console.ask("Select an option #", choices=choices, default=default)
    return None if i is "None" else int(i)



from .channel import select_from_table
from ..telem import DATA_TYPES, DataType, TIME_UNITS
from .flow import Context


def prompt_data_type_select(ctx: Context, allow_none: bool = False) -> DataType | None:
    """Prompts the user to select a data type from a list of all available data
    types.

    :param ctx: The current flow Context.
    :param allow_none: Whether to allow the user to select None.
    """
    i = select_from_table(
        ctx,
        ["option", "data_type"],
        [{"data_type": name.string()} for name in DATA_TYPES],
        allow_none=allow_none,
    )
    return DATA_TYPES[i] if i is not None else None


def prompt_time_units_select(ctx: Context, allow_none: bool = False) -> str | None:
    """Prompts the user to select a time unit from a list of all available time
    units.

    :param ctx: The current flow Context.
    :param allow_none: Whether to allow the user to select None.
    """
    opts = list(TIME_UNITS.keys())
    i = select_from_table(
        ctx,
        ["unit"],
        [{"unit": unit} for unit in opts],
        allow_none=allow_none,
    )
    return opts[i] if i is not None else None



import os
from pathlib import Path

from synnax import Synnax
from synnax.cli.connect import prompt_client_options, connect_client
from synnax.cli.console.rich import RichConsole
from synnax.cli.flow import Context
from synnax.config import ConfigFile, ClustersConfig, ClusterConfig


def login():
    """Logs the user into a Synnax cluster and saves the parameters to the configuration
    file.
    """
    ctx = Context(console=RichConsole())
    options = prompt_client_options(ctx)
    connect_client(ctx, options)
    cfg = ClustersConfig(ConfigFile(Path(os.path.expanduser("~/.synnax"))))
    cfg.set(ClusterConfig(options=options))

    cfg.get()


def credentials():
    """Prints the credentials of the currently logged in cluster."""




import fnmatch

from ..flow import Context


def prompt_group_channel_names(
    ctx: Context,
    channels: list[str],
) -> dict[str, list[str]] | None:
    ctx.console.info("""
    You can enter 'all' for all channels or a comma-separated list of:
    1) Names (e.g. 'channel1, channel2, channel3')
    2) Channel indices (e.g. '1, 2, 3')
    3) A pattern to match (e.g. 'channel*, sensor*')
    4) A combination of the above (e.g. '1, 2, channel3, my_dog*')
    """)
    res = ctx.console.ask("channels")
    return group_channel_names(ctx, channels, res.split(","))


def group_channel_names(
    ctx: Context,
    all_names: list[str],
    matchers: list[str],
):
    grouped = {}
    for entry in matchers:
        entry = entry.strip()
        channels = []
        if entry.isdigit():
            index = int(entry)
            if index < 0 or index >= len(all_names):
                ctx.console.error(f"Invalid channel index: {index}")
                continue
            channels.append(all_names[index])
        else:
            for name in all_names:
                if fnmatch.fnmatch(name, entry):
                    channels.append(name)
            if not channels:
                ctx.console.error(f"No channels match pattern: {entry}")
                if not ctx.console.confirm("Continue?", default=True):
                    return None
        grouped[entry] = channels
    return grouped



import numpy as np

from synnax.channel.payload import ChannelPayload

from .payload import SegmentPayload
from .sugared import NumpySegment


class NumpyEncoderDecoder:
    @staticmethod
    def encode(segment: NumpySegment) -> SegmentPayload:
        return SegmentPayload(
            channel_key=segment.channel.key,
            start=segment.start,
            data=segment.data.tobytes(),
        )

    @staticmethod
    def decode(channel: ChannelPayload, segment: SegmentPayload) -> NumpySegment:
        return NumpySegment(
            channel=channel,
            start=segment.start,
            data=np.frombuffer(segment.data, dtype=channel.data_type.numpy_type),
        )



from freighter import Payload

from synnax.telem import TimeStamp


class SegmentPayload(Payload):
    """A payload container that represent a segment of data exchanged to and from the
    Synnax server.
    """

    data: bytes = b""
    channel_key: str = ""
    start: TimeStamp = TimeStamp(0)



from enum import Enum

from freighter import (
    EOF,
    ExceptionPayload,
    Payload,
    Stream,
    StreamClient,
    decode_exception,
)
from numpy import ndarray

from synnax.channel.registry import ChannelRegistry
from synnax.exceptions import UnexpectedError, ValidationError, Field, GeneralError
from synnax.telem import Size, TimeStamp, UnparsedTimeStamp

from .encoder import NumpyEncoderDecoder
from .payload import SegmentPayload
from .splitter import Splitter
from .sugared import NumpySegment, SugaredBinarySegment
from .validate import ContiguityValidator, ScalarTypeValidator, Validator


class _Command(int, Enum):
    NONE = 0
    WRITE = 1
    COMMIT = 2
    ERROR = 3


class _Request(Payload):
    command: _Command
    open_keys: list[str]
    segments: list[SegmentPayload]


class _Response(Payload):
    command: _Command
    ack: bool
    error: ExceptionPayload


class CoreWriter:
    """Used to write telemetry to a set of channels in time-order. It should not be
    instantiated directly, and should instead be created using the segment client.

    Using a writer is ideal when writing large volumes of data (such as recording telemetry
    from a sensor), but it is relatively complex and challenging to use. If you're looking
    to write a contiguous block of telemetry, see the segment Client write method instead.
    """

    _ENDPOINT = "/segment/write"
    client: StreamClient
    stream: Stream[_Request, _Response] | None
    keys: list[str]

    def __init__(self, client: StreamClient) -> None:
        self.client = client

    def open(self, keys: list[str]):
        """Opens the writer, acquiring an exclusive lock on the given
        channels for the duration of the writer's lifetime. open must be called before
        any other writer methods.

        :param keys: A list of keys representing the channels the writer will write to.
        """
        self.keys = keys
        self.stream = self.client.stream(self._ENDPOINT, _Request, _Response)
        self.stream.send(_Request(command=_Command.NONE, open_keys=keys, segments=[]))
        res, exc = self.stream.receive()
        if exc is not None:
            raise exc

    def write(self, segments: list[SegmentPayload]) -> bool:
        """Validates and writes the given segments to the database. The provided segments
        must:

            1. Be in time order (on a per-channel basis).
            2. Have channel keys in the set of keys provided to open.
            3. Have non-zero length data with the correct data type for the given channel.

        :param segments: A list of segments to write to the database.
        :returns: False if the writer has accumulated an error. In this case,
        the caller should stop executing requests and close the writer.
        """
        self._assert_open()
        if self.stream.received():
            return False

        self._check_keys(segments)
        err = self.stream.send(_Request(command=_Command.WRITE, open_keys=[], segments=segments))
        if err is not None:
            raise err
        return True

    def commit(self) -> bool:
        self._assert_open()
        if self.stream.received():
            return False
        err = self.stream.send(_Request(command=_Command.COMMIT, open_keys=[], segments=[]))
        if err is not None:
            raise err

        while True:
            res, err = self.stream.receive()
            if err is not None:
                raise err
            if res.command == _Command.COMMIT:
                return res.ack

    def error(self) -> Exception:
        self.stream.send(_Request(command=_Command.ERROR, open_keys=[], segments=[]))
        res, err = self.stream.receive()
        if err is not None:
            raise err
        return decode_exception(res.error)

    def close(self):
        """Closes the writer, raising any accumulated error encountered during operation.
        A writer MUST be closed after use, and this method should probably be placed in
        a 'finally' block. If the writer is not closed, the database will not release
        the exclusive lock on the channels, preventing any other callers from writing to
        them. It also might leak resources and threads.
        """
        self._assert_open()
        self.stream.close_send()
        res, err = self.stream.receive()
        if err is None:
            err = decode_exception(res.error)
        if not isinstance(err, EOF):
            raise err

    def _check_keys(self, segments: list[SegmentPayload]):
        for segment in segments:
            if segment.channel_key not in self.keys:
                raise ValidationError(
                    Field(
                        "key",
                        f"key {segment.key} is not in the list of keys for this writer.",
                    )
                )

    def _assert_open(self):
        if self.stream is None:
            raise GeneralError(
                "Writer is not open. Please open before calling write() or close()."
            )


class NumpyWriter:
    """Used to write telemetry to a set of channels in time-order. It should not be
    instantiated directly, and should instead be created using the segment client.

    Using a writer is ideal when writing large volumes of data (such as recording telemetry
    from a sensor), but it is relatively complex and challenging to use. If you're looking
    to write a contiguous block of telemetry, see the segment Client write method instead.
    """

    core: CoreWriter
    validators: list[Validator]
    encoder: NumpyEncoderDecoder
    splitter: Splitter
    channels: ChannelRegistry

    def __init__(
        self,
        core: CoreWriter,
        channels: ChannelRegistry,
    ) -> None:
        self.core = core
        self.validators = [
            ScalarTypeValidator(),
            # ContiguityValidator(dict(), allow_no_high_water_mark=True),
        ]
        self.encoder = NumpyEncoderDecoder()
        self.splitter = Splitter(threshold=Size(4e6))
        self.channels = channels

    def open(self, keys: list[str]):
        """Opens the writer, acquiring an exclusive lock on the given
        channels for the duration of the writer's lifetime. open must be called before
        any other writer methods.
        """
        self.core.open(keys)

    def commit(self) -> bool:
        return self.core.commit()

    def error(self) -> Exception:
        return self.core.error()

    def write(self, to: str, start: UnparsedTimeStamp, data: ndarray) -> bool:
        """Writes the given telemetry to the database.

        :param to: The key of the channel to write to. This key must be present
        in the list of keys the writer was opened with.
        :param start: The start timestamp of the first sample in data. This must be
        equal to the end of the previous segment written to the channel (unless
        this is the first segment written to the channel).
        :param data: The telemetry to write. This must be a 1D numpy array with
        the same data type as the channel.
        :returns: False if the writer has accumulated an error. In this case,
        the caller should stop executing requests and close the writer.
        """
        ch = self.channels.get(to)
        seg = NumpySegment(ch, TimeStamp(start), data)
        for val in self.validators:
            val.validate(seg)
        encoded = SugaredBinarySegment.sugar(ch, self.encoder.encode(seg))
        split = self.splitter.split(encoded)
        return self.core.write([seg.payload() for seg in split])

    def close(self):
        """Closes the writer, raising any accumulated error encountered during operation.
        A writer MUST be closed after use, and this method should probably be placed in
        a 'finally' block. If the writer is not closed, the database will not release
        the exclusive lock on the channels, preventing any other callers from writing to
        them. It also might leak resources and threads.
        """
        self.core.close()



import numpy as np

from .console import RichConsole
from .flow import Context
from .io import prompt_new_reader, io_factory
from .telem import prompt_time_units_select
from ..telem import convert_time_units


def convert_timestamp_precision():
    ctx = Context(console=RichConsole())
    reader = prompt_new_reader(ctx)
    c = ctx.console.ask("Which channel would you like to convert?")
    channels = reader.channels()
    ch = next((ch for ch in channels if ch.name == c), None)
    if ch is None:
        ctx.console.error(f"Channel not found: {c}")
        return

    ctx.console.info("What is the current precision?")
    curr = prompt_time_units_select(ctx)
    ctx.console.info("What is the desired precision?")
    desired = prompt_time_units_select(ctx)
    reader.set_chunk_size(100000000)
    df = reader.read()
    converted = convert_time_units(df[ch.name], curr, desired)
    df[ch.name] = converted.astype(np.int64)
    w = io_factory.new_writer(reader.path().parent / "converted.csv")
    w.write(df)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['synnax',
 'synnax.channel',
 'synnax.cli',
 'synnax.cli.console',
 'synnax.cli.flow',
 'synnax.config',
 'synnax.ingest',
 'synnax.io',
 'synnax.segment',
 'synnax.user']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.5.0,<3.0.0',
 'keyring>=23.9.3,<24.0.0',
 'mypy>=0.971,<0.972',
 'pandas>=1.4.3,<2.0.0',
 'pick>=2.0.2,<3.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'rich>=12.6.0,<13.0.0',
 'synnax-freighter>=0.2.1,<0.3.0',
 'websockets>=10.3,<11.0']

entry_points = \
{'console_scripts': ['credentials = synnax.cli.login:credentials',
                     'login = synnax.cli.login:login',
                     'root = synnax.cli.ingest:ingestion',
                     'tsconvert = synnax.cli.util:convert_timestamp_precision']}

setup_kwargs = {
    'name': 'synnax',
    'version': '0.2.4',
    'description': 'Synnax Client Library',
    'long_description': 'None',
    'author': 'emiliano bonilla',
    'author_email': 'emilbon99@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://synnaxlabs.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from datetime import datetime, timedelta, timezone, tzinfo
from typing import Union, get_args

import numpy as np
import pandas as pd
from pydantic import BaseModel

from .exceptions import ContiguityError


class TimeStamp(int):
    """TimeStamp represents a 64 bit nanosecond-precision UTC timestamp. The TimeStamp
    constructor accepts a variety of types and will attempt to convert them to a
    TimeStamp. The following types are supported:

    * TimeStamp - returns the TimeStamp.
    * TimeSpan - treats the TimeSpan as a duration since the Unix epoch in UTC.
    * pd.TimeStamp - converts the pandas TimeStamp to a TimeStamp. If the timestamp is
    not timezone aware, it is assumed to be in the local timezone.
    * datetime - converts the datetime to a TimeStamp.  If the datetime is not timezone
    aware, it is assumed to be in the local timezone.
    * timedelta - treats the timedelta as a duration since the Unix epoch in UTC.
    * np.datetime64 - treats the numpy datetime64 as a duration since the Unix epoch in
    UTC.
    * int - treats the int as a nanosecond duration since the Unix epoch and in UTC.
    TimeStamp.

    :param value: An unparsed timestamp value that can be converted to a TimeStamp.
    """

    def __new__(cls, value: UnparsedTimeStamp, *args, **kwargs):
        if isinstance(value, TimeStamp):
            return value

        if isinstance(value, TimeSpan):
            value = int(value)
        elif isinstance(value, pd.Timestamp):
            # Convert the timestamp to a timezone aware datetime
            value = int(float(SECOND) * value.to_pydatetime().astimezone().timestamp())
        elif isinstance(value, datetime):
            value = int(float(SECOND) * value.timestamp())
        elif isinstance(value, timedelta):
            value = int(float(SECOND) * value.total_seconds())
        elif isinstance(value, np.datetime64):
            # Assume the datetime64 is in UTC
            value = int(pd.Timestamp(value).asm8.view(np.int64))
        elif isinstance(value, np.int64):
            value = int(value)
        elif isinstance(value, int):
            return super().__new__(cls, int(value))
        else:
            raise TypeError(f"Cannot convert {type(value)} to TimeStamp")

        return super().__new__(cls, value)

    def __init__(self, value: UnparsedTimeStamp, *args, **kwargs):
        pass

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        return cls(v)

    def span(self, other: UnparsedTimeStamp) -> TimeSpan:
        """Returns the TimeSpan between two timestamps. This span is guaranteed to be
        positive.
        """
        return TimeRange(self, other).make_valid().span()

    def datetime(self, tzinfo: tzinfo | None = None) -> datetime:
        """Returns the TimeStamp represented as a timezone aware datetime object.

        :param tzinfo: the timezone to use for the datetime. If None, the local timezone
        is used.
        :return: a datetime object
        """
        return (
            datetime.utcfromtimestamp(self / SECOND)
            .replace(tzinfo=timezone.utc)
            .astimezone(tzinfo)
        )

    def is_zero(self) -> bool:
        """Checks if the timestamp is the Unix epoch.
        :return: True if the TimeStamp is zero, False otherwise
        """
        return self == 0

    def after(self, ts: UnparsedTimeStamp) -> bool:
        """Returns true if the TimeStamp is after the given TimeStamp.
        :param ts: the TimeStamp to compare to
        :return: True if the TimeStamp is after the given TimeStamp, False otherwise
        """
        return super().__gt__(TimeStamp(ts))

    def after_eq(self, ts: UnparsedTimeStamp) -> bool:
        """Returns true if the TimeStamp is after or equal to the given TimeStamp.
        :param ts: the TimeStamp to compare to
        :return: True if the TimeStamp is after or equal to the given TimeStamp an d False
        otherwise.
        """
        return super().__ge__(TimeStamp(ts))

    def before(self, ts: UnparsedTimeStamp) -> bool:
        """Returns true if the TimeStamp is before the given TimeStamp.
        :param ts: the TimeStamp to compare to
        :return: True if the TimeStamp is before the given TimeStamp, False otherwise
        """
        return super().__lt__(TimeStamp(ts))

    def before_eq(self, ts: UnparsedTimeStamp) -> bool:
        """Returns true if the TimeStamp is before or equal to the given TimeStamp.
        :param ts: the TimeStamp to compare to
        :return: True if the TimeStamp is before or equal to the given TimeStamp, and False
        otherwise.
        """
        return super().__le__(TimeStamp(ts))

    def span_range(self, span: TimeSpan) -> TimeRange:
        """Returns a TimeRange that spans the given TimeSpan.
        :param span: the TimeSpan to span
        :return: a TimeRange that spans the given TimeSpan
        """
        return TimeRange(self, self + span).make_valid()

    def range(self, ts: UnparsedTimeStamp) -> TimeRange:
        """Returns a new TimeRange spanning the provided time stamps
        :param ts: the second time stamp
        :return: a new TimeRange spanning the provided time stamps
        """
        return TimeRange(self, TimeStamp(ts)).make_valid()

    def add(self, ts: UnparsedTimeStamp) -> TimeStamp:
        """Returns a new TimeStamp that is the sum of the two TimeStamps.
        :param ts: the second TimeStamp
        :return: a new TimeStamp that is the sum of the two TimeStamps
        """
        return TimeStamp(super().__add__(TimeStamp(ts)))

    def sub(self, ts: UnparsedTimeStamp) -> TimeStamp:
        """Returns a new TimeStamp that is the difference of the two TimeStamps.
        :param ts: the second TimeStamp
        :return: a new TimeStamp that is the difference of the two TimeStamps
        """
        return TimeStamp(super().__sub__(TimeStamp(ts)))

    def __add__(self, rhs: UnparsedTimeStamp) -> TimeStamp:
        return self.add(rhs)

    def __sub__(self, rhs: UnparsedTimeStamp) -> TimeStamp:
        return self.sub(rhs)

    def __lt__(self, rhs: UnparsedTimeStamp) -> bool:
        return self.before(rhs)

    def __le__(self, rhs: UnparsedTimeStamp) -> bool:
        return self.before_eq(rhs)

    def __ge__(self, rhs: UnparsedTimeStamp) -> bool:
        return self.after_eq(rhs)

    def __gt__(self, rhs: UnparsedTimeStamp) -> bool:
        return self.after(rhs)

    def __eq__(self, rhs: object) -> bool:
        if isinstance(rhs, get_args(UnparsedTimeStamp)):
            return super().__eq__(TimeStamp(rhs))
        return NotImplemented


def now() -> TimeStamp:
    """Returns the current time as a TimeStamp."""
    return TimeStamp(datetime.now())


def since(t: TimeStamp) -> TimeSpan:
    """Returns the amount of time between the given timestamp and the current time."""
    return now().span(t)


class TimeSpan(int):
    """TimeSpan represents a 64 bit nanosecond-precision duration. The TimeSpan constructor
    accepts a variety of different types and will attempt to convert them to a TimeSpan.
    The following types are supported:

    * int: the number of nanoseconds.
    * np.int64: the number of nanoseconds.
    * TimeSpan: returns a copy of the TimeSpan.
    * TimeStamp: the difference between the TimeStamp and the Unix epoch
    * timedelta: the duration of the timedelta.
    * pands.Timedelta: the duration of the Timedelta.
    * np.timedelta64: the duration of the timedelta.
    """

    def __new__(cls, value: UnparsedTimeSpan, *args, **kwargs):
        if isinstance(value, int):
            return super().__new__(cls, value)
        elif isinstance(value, TimeSpan):
            return value

        if isinstance(value, timedelta):
            value = int(float(SECOND) * value.total_seconds())
        elif isinstance(value, pd.Timedelta):
            value = int(float(SECOND) * value.total_seconds())
        elif isinstance(value, np.timedelta64):
            value = int(float(SECOND) * pd.Timedelta(value).total_seconds())
        elif isinstance(value, np.int64):
            value = int(value)
        else:
            raise TypeError(f"Cannot convert {type(value)} to TimeSpan")

        return super().__new__(cls, value)

    def __init__(self, value: UnparsedTimeSpan, *args, **kwargs):
        pass

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        return cls(value)

    def __repr__(self) -> str:
        return f"TimeSpan({super().__repr__()})"

    def delta(self) -> timedelta:
        """Returns the TimeSpan represented as a timedelta object.
        :return: a timedelta object
        """
        return timedelta(seconds=self.seconds())

    def seconds(self) -> float:
        """Returns the TimeSpan represented as a number of seconds.
        :return: a number of seconds
        """
        return float(self / SECOND)

    def is_zero(self) -> bool:
        """Returns true if the TimeSpan is zero.
        :return: True if the TimeSpan is zero, False otherwise
        """
        return self == 0

    def add(self, ts: UnparsedTimeSpan) -> TimeSpan:
        """Returns a new TimeSpan that is the sum of the two TimeSpans.
        :param ts: the second TimeSpan
        :return: a new TimeSpan that is the sum of the two TimeSpans
        """
        return TimeSpan(super().__add__(TimeSpan(ts)))

    def sub(self, ts: UnparsedTimeSpan) -> TimeSpan:
        """Returns a new TimeSpan that is the difference of the two TimeSpans.
        :param ts: the second TimeSpan
        :return: a new TimeSpan that is the difference of the two TimeSpans
        """
        return TimeSpan(super().__sub__(TimeSpan(ts)))

    def __add__(self, other: UnparsedTimeSpan) -> TimeSpan:
        return self.add(other)

    def __sub__(self, other: UnparsedTimeSpan) -> TimeSpan:
        return self.sub(other)

    def __mul__(self, other: UnparsedTimeSpan) -> TimeSpan:
        return TimeSpan(super().__mul__(TimeSpan(other)))

    def __rmul__(self, other: UnparsedTimeSpan) -> TimeSpan:
        return self.__mul__(other)

    def __gt__(self, other: UnparsedTimeSpan) -> bool:
        return super().__gt__(TimeSpan(other))

    def __ge__(self, other: UnparsedTimeSpan) -> bool:
        return super().__ge__(TimeSpan(other))

    def __lt__(self, other: UnparsedTimeSpan) -> bool:
        return super().__lt__(TimeSpan(other))

    def __le__(self, other: UnparsedTimeSpan) -> bool:
        return super().__le__(TimeSpan(other))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, get_args(UnparsedTimeSpan)):
            return NotImplemented
        return super().__eq__(TimeSpan(other))


TIME_STAMP_MIN = TimeStamp(0)
TIME_STAMP_MAX = TimeStamp(2 ** 63 - 1)
NANOSECOND = TimeSpan(1)
NANOSECOND_UNITS = "ns"
MICROSECOND = TimeSpan(1000) * NANOSECOND
MICROSECOND_UNITS = "us"
MILLISECOND = TimeSpan(1000) * MICROSECOND
MILLISECOND_UNITS = "ms"
SECOND = TimeSpan(1000) * MILLISECOND
SECOND_UNITS = "s"
MINUTE = TimeSpan(60) * SECOND
MINUTE_UNITS = "m"
HOUR = TimeSpan(60) * MINUTE
HOUR_UNITS = "h"
TIME_SPAN_MAX = TimeSpan(0xFFFFFFFFFFFFFFFF)

TIME_UNITS = {
    NANOSECOND_UNITS: NANOSECOND,
    MICROSECOND_UNITS: MICROSECOND,
    MILLISECOND_UNITS: MILLISECOND,
    SECOND_UNITS: SECOND,
    MINUTE_UNITS: MINUTE,
    HOUR_UNITS: HOUR,
}


def convert_time_units(data: np.ndarray, _from: str, to: str):
    """Converts the data from one time unit to another.
    :param data: the data to convert
    :param _from: the units of the data
    :param to: the units to convert to
    :return: the data in the new units
    """
    if _from == to:
        return data
    return data * TIME_UNITS[_from] / TIME_UNITS[to]


class Rate(float):
    """Rate represents a data rate in Hz"""

    def __new__(cls, value: UnparsedRate):
        if isinstance(value, float):
            return super().__new__(cls, value)
        if isinstance(value, Rate):
            return value
        if isinstance(value, TimeSpan):
            value = 1 / value.seconds()
        elif isinstance(value, int):
            value = float(value)
        else:
            raise TypeError(f"Cannot convert {type(value)} to Rate")
        return super().__new__(cls, value)

    def __init__(self, value: UnparsedRate):
        pass

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        return cls(v)

    def period(self) -> TimeSpan:
        """Returns the period of the rate as a TimeSpan"""
        return TimeSpan(int(1 / self * float(SECOND)))

    def sample_count(self, time_span: UnparsedTimeSpan) -> int:
        """Returns the number of samples in the given TimeSpan at this rate"""
        return int(TimeSpan(time_span).seconds() * self)

    def byte_size(self, time_span: UnparsedTimeSpan, density: Density) -> Size:
        """Calculates the amount of bytes occupied by the given TimeSpan at the given
        rate and sample density."""
        return Size(self.sample_count(time_span) * int(density))

    def span(self, sample_count: int) -> TimeSpan:
        """Returns the TimeSpan that corresponds to the given number of samples at this
        rate."""
        return self.period() * sample_count

    def size_span(self, size: Size, density: Density) -> TimeSpan:
        """Returns the TimeSpan that corresponds to the given number of bytes at this
        rate and sample density."""
        if size % density != 0:
            raise ContiguityError(f"Size {size} is not a multiple of density {density}")
        return self.span(int(size / density))

    def __str__(self):
        return str(int(self)) + "Hz"

    def __repr__(self):
        return f"Rate({super().__repr__()} Hz)"


HZ = Rate(1)
KHZ = Rate(1000) * HZ
MHZ = Rate(1000) * KHZ


class TimeRange(BaseModel):
    start: TimeStamp
    end: TimeStamp

    def __init__(self, start: UnparsedTimeStamp, end: UnparsedTimeStamp):
        super().__init__(start=TimeStamp(start), end=TimeStamp(end))

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, TimeRange):
            return cls(v.start, v.end)
        return cls(start=v[0], end=v[1])

    def span(self) -> TimeSpan:
        return TimeSpan(self.end - self.start)

    def make_valid(self) -> TimeRange:
        if not self.is_valid():
            return self.swap()
        return self

    def is_zero(self) -> bool:
        return self.span().is_zero()

    def bound_by(self, otr: TimeRange) -> TimeRange:
        if otr.start.after(self.start):
            self.start = otr.start
        if otr.start.after(self.end):
            self.end = otr.start
        if otr.end.before(self.end):
            self.end = otr.end
        if otr.end.before(self.start):
            self.start = otr.end
        return self

    def contains_stamp(self, ts: TimeStamp) -> bool:
        return ts.after_eq(self.start) and ts.before(self.end)

    def contains_range(self, tr: TimeRange) -> bool:
        return self.start.before_eq(tr.start) and self.end.after_eq(tr.end)

    def overlaps_with(self, tr: TimeRange) -> bool:
        return (
            self.contains_stamp(tr.start)
            or self.contains_stamp(tr.end)
            or tr.contains_range(self)
        )

    def swap(self) -> TimeRange:
        self.start, self.end = self.end, self.start
        return self

    def is_valid(self) -> bool:
        return self.span() >= TimeSpan(0)

    def __str__(self):
        return str(self.start) + " - " + str(self.end)


class Density(int):
    def __new__(cls, value: UnparsedDensity):
        if isinstance(value, Density):
            return value
        if isinstance(value, int):
            return super().__new__(cls, value)
        raise TypeError(f"Cannot convert {type(value)} to Density")

    def __init__(self, value):
        pass

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        return cls(v)

    def __repr__(self):
        return f"Density({super().__repr__()})"


class Size(int):
    def __str__(self):
        return super(Size, self).__str__() + "B"


BYTE = Size(1)
KILOBYTE = Size(1024) * BYTE
MEGABYTE = Size(1024) * KILOBYTE
GIGABYTE = Size(1024) * MEGABYTE

TIME_RANGE_MAX = TimeRange(TIME_STAMP_MIN, TIME_STAMP_MAX)
DENSITY_UNKNOWN = Density(0)
BIT64 = Density(8)
BIT32 = Density(4)
BIT16 = Density(2)
BIT8 = Density(1)


class DataType(str):
    """DataType represents a data type as a string"""

    def __new__(cls, value: UnparsedDataType):
        if isinstance(value, DataType):
            return value
        if isinstance(value, str):
            return super().__new__(cls, value)
        try:
            if issubclass(value, np.ScalarType):
                value = NUMPY_TO_DATA_TYPE.get(value, None)
                if value is None:
                    raise TypeError(f"Cannot convert {value} to DataType")
                return value
        except TypeError:
            pass
        raise TypeError(f"Cannot convert {type(value)} to DataType")

    def __init__(self, value: UnparsedDataType):
        pass

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        return cls(v)

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(type="string")

    @property
    def numpy_type(self, _raise: bool = False) -> np.ScalarType | None:
        """Converts a built-in DataType to a numpy type Scalar Type
        :param _raise: If True, raises a TypeError if the DataType is not a numpy type.
        :return: The numpy type
        """
        npt = DATA_TYPE_TO_NUMPY.get(self, None)
        if npt is None and _raise:
            raise TypeError(f"Cannot convert {self} to numpy type")
        return npt

    def __repr__(self):
        return f"DataType({super(DataType, self).__repr__()})"

    def string(self):
        return str(self)


DATA_TYPE_UNKNOWN = DataType("")
FLOAT64 = DataType("float64")
FLOAT32 = DataType("float32")
INT64 = DataType("int64")
INT32 = DataType("int32")
INT16 = DataType("int16")
INT8 = DataType("int8")
UINT64 = DataType("uint64")
UINT32 = DataType("uint32")
UINT16 = DataType("uint16")
UINT8 = DataType("uint8")
DATA_TYPES = [
    FLOAT64,
    FLOAT32,
    INT64,
    INT32,
    INT16,
    INT8,
    UINT64,
    UINT32,
    UINT16,
    UINT8,
]

UnparsedTimeStamp = Union[
    int,
    TimeStamp,
    TimeSpan,
    datetime,
    timedelta,
    np.datetime64,
    np.int64,
]
UnparsedTimeSpan = Union[
    int,
    TimeSpan,
    TimeStamp,
    timedelta,
    np.timedelta64,
]
UnparsedRate = Union[int, float, TimeSpan, Rate]
UnparsedDensity = Density | int
UnparsedDataType = (*np.ScalarType, DataType, str)

DATA_TYPE_TO_NUMPY: dict[str, np.ScalarType] = {
    FLOAT64: np.float64,
    FLOAT32: np.float32,
    INT64: np.int64,
    INT32: np.int32,
    INT16: np.int16,
    INT8: np.int8,
    UINT64: np.uint64,
    UINT32: np.uint32,
    UINT16: np.uint16,
    UINT8: np.uint8,
}
NUMPY_TO_DATA_TYPE = {v: k for k, v in DATA_TYPE_TO_NUMPY.items()}



import synnax
from .channel import (
    maybe_select_channel,
    channel_name_table,
    Channel,
)
from .connect import connect_client, prompt_client_options, load_config_options
from .console.rich import RichConsole
from .io import prompt_file
from .telem import prompt_data_type_select
from .. import Synnax
from .flow import Context, Flow
from ..ingest.row import RowIngestionEngine
from ..io import RowReader, ReaderType, IOFactory, ChannelMeta
from ..telem import INT64, Rate
from ..cli.console.channel import prompt_group_channel_names
from ..exceptions import QueryError


class IngestionCLI:
    factory: IOFactory
    reader: RowReader | None
    client: Synnax | None
    filtered_channels: list[ChannelMeta] | None
    not_found: list[ChannelMeta] | None
    db_channels: list[synnax.Channel] | None

    def __init__(self, factory: IOFactory):
        self.factory = factory
        self.reader = None
        self.client = None
        self.filtered_channels = None
        self.not_found = None
        self.db_channels = None


def run_ingestion(ctx: Context, cli: IngestionCLI) -> None:
    """Runs the ingestion process.
    """
    if cli.reader.type() == ReaderType.Row:
        engine = RowIngestionEngine(cli.client, cli.reader, cli.db_channels)
    else:
        raise NotImplementedError("Only row ingestion is supported at this time.")
    ctx.console.info("Starting ingestion process...")
    engine.run()


def ingestion():
    flow = Flow(Context(console=RichConsole()))
    flow.add("initialize_reader", initialize_reader)
    flow.add("connect_client", _connect_client)
    flow.add("ingest_all", ingest_all)
    flow.add("channels_to_ingest", channels_to_ingest)
    flow.add("validate_channels", validate_channels)
    flow.add("create_channels", create_channels)
    flow.add("ingest", run_ingestion)
    flow.run(IngestionCLI(synnax.io.IOFactory()), "initialize_reader")


def initialize_reader(
    ctx: Context,
    cli: IngestionCLI,
) -> str | None:
    ctx.console.info("Welcome to the Synnax ingestion CLI! Let's get started.")
    ctx.console.info("Please select a file to ingest.")
    path = prompt_file(ctx)
    if path is None:
        return None
    cli.reader = cli.factory.new_reader(path)
    return "connect_client"


def _connect_client(ctx: Context, cli: IngestionCLI) -> str | None:
    """Prompts the user to connect to a Synnax client.
    """
    opts = load_config_options(ctx)
    if opts is None:
        opts = prompt_client_options(ctx)
    else:
        ctx.console.info("Using saved credentials.")
    cli.client = connect_client(ctx, opts)
    return "ingest_all" if cli.client else None


def ingest_all(ctx: Context, cli: IngestionCLI) -> str | None:
    """Prompts the user to ingest all channels.
    """
    if ctx.console.confirm("Would you like to ingest all channels?", default=True):
        cli.filtered_channels = cli.reader.channels()
        return "validate_channels"
    else:
        return "channels_to_ingest"


def channels_to_ingest(ctx: Context, cli: IngestionCLI) -> str | None:
    """Prompts the user to select channels to ingest.
    """
    ctx.console.info("Which channels would you like to ingest?")
    channels = cli.reader.channels()
    grouped = prompt_group_channel_names(ctx, [ch.name for ch in channels])
    if grouped is None or len(grouped) == 0:
        return None
    all_names = [v for l in grouped.values() for v in l]
    cli.filtered_channels = [ch for ch in channels if ch.name in all_names]
    return "validate_channels"


def validate_channels(ctx: Context, cli: IngestionCLI) -> str | None:
    """Validates that all channels in the file exist in the database. If not, prompts
    the user to create them.
    """
    ctx.console.info("Validating channels in file...")
    cli.not_found = []
    cli.db_channels = []
    for channel in cli.filtered_channels:
        try:
            ch = maybe_select_channel(
                ctx,
                cli.client.channel.filter(names=[channel.name]),
                channel.name,
            )
        except QueryError:
            cli.not_found.append(channel)
        else:
            cli.db_channels.append(ch)

    if len(cli.not_found) > 0:
        ctx.console.info("The following channels were not found in the database:")
        channel_name_table(ctx, [ch.name for ch in cli.not_found])
        if not ctx.console.confirm("Would you like to create them?"):
            return None
        return "create_channels"

    return "ingest"


def create_indexes(
    ctx: Context,
    cli: IngestionCLI,
    options: list[ChannelMeta],
) -> list[ChannelMeta]:
    """Prompts the user to create index channels.
    """
    grouped = prompt_group_channel_names(ctx, [ch.name for ch in options])
    names = [name for v in grouped.values() for name in v]
    for name in names:
        ch = cli.client.channel.create(name=name, is_index=True, data_type=INT64)
        cli.db_channels.append(ch)
    return [ch for ch in options if ch.name not in names]


def group_by_idx(
    ctx: Context,
    options: list[ChannelMeta],
) -> dict[str, list[ChannelMeta]] | None:
    """Prompts the user to group channels by their index/rate
    """
    if not ctx.console.confirm(
        "Do all non-indexed channels have the same data rate or index?"
    ):
        if not ctx.console.confirm("Can you group channels by data rate or index?"):
            return {v.name: v for v in options}
        grouped = prompt_group_channel_names(ctx, [ch.name for ch in options])
        if grouped is None or len(grouped) == 0:
            return None
        return {k: v for k, v in grouped.items()}
    return {"__all__": options}


def create_channels(ctx: Context, cli: IngestionCLI) -> str | None:
    if ctx.console.confirm("Are any channels indexed (e.g. timestamps)?"):
        cli.not_found = create_indexes(ctx, cli, cli.not_found)

    idx_grouped = group_by_idx(ctx, cli.not_found)
    if idx_grouped is None:
        return None

    # First thing we must try to do is get the set of all data rates
    # or indexes that are being used in db_channels

    assigned_dr_idx = {}
    for key, group in idx_grouped.items():
        if key != "__all__":
            ctx.console.info(f"Assigning data rate or index to {key}")
        _choice = ctx.console.ask("Enter the name of an index or a data rate")
        if _choice.replace(".", "").isdigit():
            assigned_dr_idx[Rate(float(_choice))] = group
        else:
            # if the user entered a string, we have an index channel, and we
            #  need to make sure that the string is a valid index. Look first
            # in not_found then try to query
            res = cli.client.channel.filter(names=[_choice])
            idx = maybe_select_channel(ctx, res, _choice)
            if not idx:
                ctx.console.ask(f"No channel found for index {_choice}")
                return None
            assigned_dr_idx[idx.key] = group

    dt_grouped = {}
    if not ctx.console.confirm("Do all of these channels have the same data type?"):
        if ctx.console.confirm("Can you group them by data type?"):
            grouped = prompt_group_channel_names(ctx, [ch.name for ch in cli.not_found])
            if grouped is None or len(grouped) == 0:
                return None
            dt_grouped = {k: v for k, v in grouped.items()}
        else:
            dt_grouped = {v: v for v in cli.not_found}
    else:
        dt_grouped["__all__"] = cli.not_found

    assigned_dts = {}
    for key, group in dt_grouped.items():
        if key != "__all__":
            ctx.console.info(f"Assigning data type to {key}")
        dt = prompt_data_type_select(ctx)
        assigned_dts[dt] = group

    to_create = []
    for rate_or_index, channels in assigned_dr_idx.items():
        # if the rate_or_index is a string, it is an index channel
        index = ""
        rate: Rate = Rate(0)
        if isinstance(rate_or_index, Rate):
            rate = rate_or_index
        else:
            index = rate_or_index

        for channel in channels:
            dt = None
            for k, v in assigned_dts.items():
                if channel in v:
                    dt = k
                    break

            to_create.append(Channel(
                name=channel.name,
                is_index=False,
                index=index,
                rate=rate,
                data_type=dt
            ))

    cli.db_channels.extend(cli.client.channel.create_many(to_create))

    return "ingest"



from freighter import HTTPClientFactory, Payload, UnaryClient

from synnax.telem import DATA_TYPE_UNKNOWN, Rate, UnparsedDataType, UnparsedRate

from .payload import ChannelPayload


class _Request(Payload):
    channels: list[ChannelPayload]


class _Response(Payload):
    channels: list[ChannelPayload]


class ChannelCreator:
    _ENDPOINT = "/channel/create"
    client: UnaryClient

    def __init__(self, client: HTTPClientFactory):
        self.client = client.post_client()

    def create(
        self,
        name: str = "",
        node_id: int = 0,
        rate: UnparsedRate = Rate(0),
        data_type: UnparsedDataType = DATA_TYPE_UNKNOWN,
        index: str = "",
        is_index: bool = False,
    ) -> ChannelPayload:
        print(data_type)
        return self.create_many([ChannelPayload(
            data_type=data_type,
            name=name,
            node_id=node_id,
            rate=rate,
            index=index,
            is_index=is_index,
        )]
        )[0]

    def create_many(self, channels: list[ChannelPayload]) -> list[ChannelPayload]:
        req = _Request(channels=channels)
        res, exc = self.client.send(self._ENDPOINT, req, _Response)
        if exc is not None:
            raise exc
        assert res is not None
        return res.channels
